<?php

/***
Alternative PHP client that does not use HttpRequest.
This still uses the cUrl module.
***/

require ('config.php');

// TODO: for security, this should be a path to a file outside the website.
define ("CONFIG_DOC", "config.xml");
define('PINSAFE_COOKIE', 'PINSAFE_AUTH');

function getPINsafeAgent() {
	$config = new Config();
	$config->load(CONFIG_DOC);
	return new PINsafe($config->server, $config->port, $config->context, $config->secret, $config->ssl, $config->ignoreCertErrors, $config->cookieSecret, $config->cookieTimeout);
}

class PINsafe {
	// PINsafe server
	private $server;

	// PINsafe port
	private $port;

	// PINsafe context
	private $context;

	// PINsafe agent secret
	private $secret;

	// SSL flag
	private $ssl;
	
	// Ignore certificate errors flag
	private $ignoreCertErrors;
	
	// Cookie secret
	private $cookieSecret;

	// Cookie timeout
	private $cookieTimeout;
	
	private $responseCode;
	
	private $warning;

	public function __construct($server, $port, $context, $secret, $ssl, $ignoreCertErrors, $cookieSecret, $cookieTimeout) {
		$this->server = $server;
		$this->port = $port;
		$this->context = $context;
		$this->secret = $secret;
		$this->ssl = $ssl;
		$this->ignoreCertErrors = $ignoreCertErrors;
		$this->cookieSecret = $cookieSecret;
		$this->cookieTimeout = $cookieTimeout;
	}

	private function getProtocol() {
		if ($this->ssl) {
			return "https://";
		}
		return "http://";
	}

	public function getCookieTimeoutSecs() {
		return $this->cookieTimeout * 60;
	}
	
	private function getUrl() {
		return $this->getProtocol() . "$this->server:$this->port/$this->context";
	}

	private function getAgentUrl() {
		return $this->getUrl() . '/AgentXML';
	}

	private function getImageUrl() {
		return $this->getUrl() . '/SCImage';
	}
	
	private function checkCertErrors($req) {
		if ($this->ssl && $this->ignoreCertErrors) {
			curl_setopt($req, CURLOPT_SSL_VERIFYPEER, false);
			curl_setopt($req, CURLOPT_SSL_VERIFYHOST, false);
		}
	}

	private function newRequest($type) {
		$doc = new DOMDocument('1.0', 'UTF-8');
		$elem = $doc->createElement('SASRequest');
		$doc->appendChild($elem);
		$this->addParameter($doc, 'Action', $type);
		$this->addParameter($doc, 'Secret', $this->secret);
		$this->addParameter($doc, 'Version', '1.0');
		return $doc;
	}

	private function addParameter($doc, $name, $value) {
	  $elem = $doc->createElement($name, $value);
	  $doc->documentElement->appendChild($elem);
  }

	private function sendRequest($doc) {
		$req = curl_init($this->getAgentUrl());
		curl_setopt($req, CURLOPT_POST, TRUE);
		curl_setopt($req, CURLOPT_HEADER, FALSE);
		$this->checkCertErrors($req);
		$reqXml = $doc->saveXML();
		curl_setopt($req, CURLOPT_HTTPHEADER, array('Content-Type: text/xml', 'Content-Encoding: UTF-8', 'Content-Length: ' . strlen($reqXml)));
		curl_setopt($req, CURLOPT_RETURNTRANSFER, TRUE);
		$fp = fopen('data://text/plain,' . $reqXml, 'r');
		curl_setopt($req, CURLOPT_INFILE, $fp);
		$result = curl_exec($req);
		curl_close($req);
		fclose($fp);
		return $result;
	}

	public function getResponseCode() {
		return $this->responseCode;
	}
	
	public function getWarning() {
		return $this->warning;
	}

	public function startSession($username) {
		$doc = $this->newRequest('sessionstart');
		$this->addParameter($doc, 'username', $username);
		$response = $this->sendRequest($doc);
		if ($response) {
			$respDoc = new DOMDocument();
			$respDoc->loadXML($response);
			$sessionIDs = $respDoc->getElementsByTagName('SessionID');
			if ($sessionIDs->length > 0) {
				return $sessionIDs->item(0)->textContent;
			}
		}
		return 0;
	}

	private function sendRequestAndTest($doc) {
		$response = $this->sendRequest($doc);
		if ($response) {
			$respDoc = new DOMDocument();
			$respDoc->loadXML($response);
			$warnings = $respDoc->getElementsByTagName('Warning');
			$this->warning = '';
			if ($warnings->length > 0) {
				$this->warning = $warnings->item(0)->textContent;
			}
			$results = $respDoc->getElementsByTagName('Result');
			if ($results->length > 0) {
				return ($results->item(0)->textContent == 'PASS');
			}
		}
		return false;
	}

	public function login($username, $password, $otc) {
		$doc = $this->newRequest('login');
		$this->addParameter($doc, 'Username', $username);
		$this->addParameter($doc, 'Password', $password);
		$this->addParameter($doc, 'OTC', $otc);
		return $this->sendRequestAndTest($doc);
	}

	public function changePIN($username, $oldPassword, $oldOTC, $newPassword, $newOTC) {
		$doc = $this->newRequest('changepin');
		$this->addParameter($doc, 'Username', $username);
		$this->addParameter($doc, 'Password', $oldPassword);
		$this->addParameter($doc, 'NewPassword', $newPassword);
		$this->addParameter($doc, 'OTC', $oldOTC);
		$this->addParameter($doc, 'NewOTC', $newOTC);
		return $this->sendRequestAndTest($doc);
	}

	public function reset($username, $resetCode) {
		$doc = $this->newRequest('reset');
		$this->addParameter($doc, 'Username', $username);
		$this->addParameter($doc, 'ResetCode', $resetCode);
		return $this->sendRequestAndTest($doc);
	}

	public function requestResetCode($username) {
		$doc = $this->newRequest('resetcode');
		$this->addParameter($doc, "Username", $username);
		return $this->sendRequestAndTest($doc);
	}
	
	public function userExists($username) {
		$doc = $this->newRequest('exists');
		$this->addParameter($doc, "Username", $username);
		return $this->sendRequestAndTest($doc);
	}

	public function getImageForSession($sessionID) {
		$req = curl_init($this->getImageUrl() . "?sessionid=$sessionID");
		curl_setopt($req, CURLOPT_HEADER, FALSE);
		curl_setopt($req, CURLOPT_RETURNTRANSFER, TRUE);
		curl_setopt($req, CURLOPT_BINARYTRANSFER, TRUE);
		$this->checkCertErrors($req);
		$result = curl_exec($req);
		curl_close($req);
		return $result;
	}

	public function getImageForUser($userID) {
		$url = $this->getImageUrl() . "?username=$userID";
		$req = curl_init($url);
		curl_setopt($req, CURLOPT_HEADER, FALSE);
		curl_setopt($req, CURLOPT_RETURNTRANSFER, TRUE);
		curl_setopt($req, CURLOPT_BINARYTRANSFER, TRUE);
		$this->checkCertErrors($req);
		$result = curl_exec($req);
		if (is_bool($result)) {
			$this->warning = curl_getinfo($req, CURLINFO_HTTP_CODE);
		}
		curl_close($req);
		return $result;
	}
	
	private function getHashValue($data) {
		$hash = crypt($data, '$6$' . $this->cookieSecret);
		return $hash;
	}
	
	public function authenticate($username, $password, $otc) {
		if ($this->login($username, $password, $otc)) {
			return $this->getAuthCookie($username);
		}
		return null;
	}
	
	public function getAuthCookie($userID) {
		if ($userID == null || $userID == '') {
			return null;
		}
		$now = new DateTime(NULL, new DateTimeZone('UTC'));
		$startDate = new DateTime('2000-01-01', new DateTimeZone('UTC'));
		$endTime = $now->getTimestamp() - $startDate->getTimestamp();
        $expirySpan = $endTime + $this->getCookieTimeoutSecs();
        $expirySecs = str_pad($expirySpan, 10, '0', STR_PAD_LEFT);
        $userID = strtolower($userID);

        return $this->getHashValue($expirySecs . $userID) . $userID . str_pad(strlen($userID), 3, '0', STR_PAD_LEFT) . $expirySecs;
	}
	
	public function getCookieUsername($cookieData) {
		$cookieLen = ($cookieData == null) ? 0 : strlen($cookieData);
		if ($cookieLen <= 16) {
			return null;
		}
		$nameStart = strlen($cookieData) - 13;
		$sNameLength = substr($cookieData, $nameStart, 3);
		$nameLength = intval($sNameLength);
		if ($nameLength <= 0) {
			return null;
		}
		if ($cookieLen <= $nameLength + 13) {
			return null;
		}
		$nameStart -= $nameLength;
		return substr($cookieData, $nameStart, $nameLength);
	}
	
	public function isCookieValid($cookieData) {
		// Split the cookie up.
		$cookieLen = ($cookieData == null) ? 0 : strlen($cookieData);
		if ($cookieLen <= 16) {
			$this->warning = 'Too short: ' . $cookieLen;
			return false;
		}
		$hashLen = $cookieLen - 10;
		$expiryString = substr($cookieData, $hashLen);
		$expiryLong = intval($expiryString);
		if ($expiryLong <= 0) {
			$this->warning = 'Invalid expiry: ' . $expiryString;
			return false;
		}
		$hashLen -= 3;
		$sNameLength = substr($cookieData, $hashLen, 3);
		$nameLength = intval($sNameLength);
		if ($nameLength <= 0) {
			$this->warning = 'Invalid name length: ' . $sNameLength;
			return false;
		}
		$hashLen -= $nameLength;
		if ($hashLen <= 0) {
			$this->warning = 'Too short (2): ' . $cookieLen;
			return false;
		}
		$username = substr($cookieData, $hashLen, $nameLength);
		$now = new DateTime(NULL, new DateTimeZone('UTC'));
		$startTime = new DateTime('2000-01-01', new DateTimeZone('UTC'));
		$expiryInt = new DateInterval("PT{$expiryLong}S");
		$expiryTime = $startTime->add($expiryInt);
		if ($expiryTime < $now) {
			$this->warning = 'Cookie expired';
			return false; // Cookie has expired.
		}
		$orgHash = substr($cookieData, 0, $hashLen);
		$timeAndUser = $expiryString . strtolower($username);
		if ($orgHash == $this->getHashValue($timeAndUser)) {
			return true;
		}
		$this->warning = 'Mismatched hash';
		return false;
	}
}

?>